/* Sample MPI code for using BLOPEX.
 *
 * To build:
 *	mpicc -O2 -Wall -c sample-mpi.c
 *	mpif77 -o sample-mpi sample-mpi.o -lBLOPEX -llapack -lblas
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <mpi.h>

#if 0
#include "lobpcg.h"
#include "multivector.h"
#include "interpreter.h"
#else
#include <blopex.h>
#endif

/* Lapack functions */
extern int dsygv_(int *itype, char *jobz, char *uplo, int *
                    n, double *a, int *lda, double *b, int *ldb,
                    double *w, double *work, int *lwork, int *info);
extern int dpotrf_(char *uplo, int *n, double *a, int * lda, int *info);

static int nglobal = 200, neig = 5;
static int nlocal, myid, nprocs, prev, next;

/* function computing y := A x */
static void
matveca(double *x, double *y)
{
  int i;
  double a = 0, b = 0;
  MPI_Status status;

  assert(nlocal > 1);

  MPI_Sendrecv(x, 1, MPI_DOUBLE, prev, 111, &a, 1, MPI_DOUBLE, prev, 111,
		MPI_COMM_WORLD, &status);
  MPI_Sendrecv(x+nlocal-1, 1, MPI_DOUBLE, next, 111, &b, 1, MPI_DOUBLE, next,
		111, MPI_COMM_WORLD, &status);

  *(y++) = -a + 2. * x[0] - x[1];
  for (i = 1; i < nlocal - 1; i++, x++)
	*(y++) = -x[0] + 2. * x[1] - x[2];
  *(y++) = -x[0] + 2. * x[1] - b;
}

/* function computing y := B x */
static void
matvecb(double *x, double *y)
{
  int i;
  double d = 1.0 / ((nglobal + 1) * (nglobal + 1));

  for (i = 0; i < nlocal; i++)
	*(y++) = d * *(x++);
}

/*------------------------ Interface functions ------------------------*/

#if 0
# define CHECK(x, y) \
  check(__func__, nlocal,  x->nact, x->list, x->data, y->nact, y->list, y->data)
static void
check(const char *func, int nlocal,  int nx, int *xlist, double *xdata,
				int ny, int *ylist, double *ydata)
{
  double sx = 0., sy = 0., *p;
  int i, j;

  for (i = 0; i < nx; i++) {
    p = xdata + xlist[i] * nlocal;
    for (j = 0; j < nlocal; j++) {
      sx += *(p++);
    }
  }
  for (i = 0; i < ny; i++) {
    p = ydata + ylist[i] * nlocal;
    for (j = 0; j < nlocal; j++) {
      sy += *(p++);
    }
  }
  printf("%-25s nx=%d, %12.5le, ny=%d, %12.5le\n", func, nx, sx, ny, sy);
}
#else
# define CHECK(x, y)
#endif

typedef struct {
    double	*data;
    int		*list;
    int		nvec, nact;	/* number of total/active vectors */
} VEC;

static void *
CreateMultiVector(int nvec)
{
  VEC *x0 = malloc(sizeof(*x0));
  int i;

  x0->nact = x0->nvec = nvec;
  x0->data = malloc(nvec * nlocal * sizeof(*x0->data));
  x0->list = malloc(nvec * sizeof(*x0->list));
  for (i = 0; i < nvec; i++)
    x0->list[i] = i;
  return x0;
}

static void *
CreateCopyMultiVector(void *src_, int copyValues)
{
   VEC *src = src_;
   VEC *dest;
   int i, nvec;
  
   dest = malloc(sizeof(*dest));
   dest->nact = dest->nvec = nvec = src->nvec;
   dest->data = malloc(nvec * nlocal * sizeof(*dest->data));
   dest->list = malloc(nvec * sizeof(*dest->list));
   for (i = 0; i < nvec; i++)
	dest->list[i] = i;
   if (copyValues)
      memcpy(dest->data, src->data, nvec * nlocal * sizeof(*dest));

   return dest;
}

static void
DestroyMultiVector(void *vec_)
{
  VEC *vec = vec_;

  free(vec->data);
  free(vec->list);
  free(vec);
}

static int
MultiVectorWidth(void *v)
{
  return ((VEC *)v)->nvec;
}

static void
MultiSetMask(void *vec_, int *mask)
{
  VEC *vec = vec_;
  int i, m = 0;

  for (i = 0; i < vec->nvec; i++)
    if (mask == NULL || mask[i] != 0)
      vec->list[m++] = i;
  vec->nact = m;
}

static void
CopyMultiVector(void *x_, void *y_)
{
  VEC *x = x_, *y = y_;
  int i;

  assert(x->nact == y->nact);
  for (i = 0; i < x->nact; i++)
    memcpy(y->data + y->list[i] * nlocal, x->data + x->list[i] * nlocal,
		nlocal * sizeof(*x->data));
  CHECK(x, y);
}

static void
ClearMultiVector(void *x_)
{
  VEC *x = x_;
  double *p;
  int i, j;

  for (i = 0; i < x->nact; i++) {
    p = x->data + x->list[i] * nlocal;
    for (j = 0; j < nlocal; j++)
      *(p++) = 0.0;
  }
}

static void
SetMultiVectorRandomValues(void *v_, int seed)
{
  VEC *v = v_;
  double *start, *end;
  int i;

  srand48(seed + myid);
  for (i = 0; i < v->nact; i++) {
    start = v->data + v->list[i] * nlocal;
    end = start + nlocal;
    while (start < end) {
       *(start++) = 2.0 * drand48() - 1.0;
    }
  }
}

static void
MultiInnerProd(void *x_, void *y_, int gh, int h, int w, double *v)
{
  VEC *x = x_, *y = y_;
  int i, j, k;
  double d, *p, *q, *r, *v0;

  assert(x->nact == h && y->nact == w);
  r = v0 = malloc(h * w * sizeof(*v0));
  for (j = 0; j < w; j++) {
    p = y->data + y->list[j] * nlocal;
    for (i = 0; i < h; i++) {
      q = x->data + x->list[i] * nlocal;
      d = 0.0;
      for (k = 0; k < nlocal; k++)
	d += p[k] * *(q++);
      *(r++) = d;
    }
  }

  if (nprocs > 1) {
    r = malloc(h * w * sizeof(*r));
    MPI_Allreduce(v0, r, w * h, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    free(v0);
    v0 = r;
  }
  for (j = 0, r = v0; j < w; j++, r += h, v += gh)
    memcpy(v, r, h * sizeof(*r));
  free(v0);
  CHECK(x, y);
}

static void
MultiInnerProdDiag(void *x_, void *y_, int *mask, int n0, double *diag)
{
  VEC *x = x_, *y = y_;
  int *list = malloc(n0 * sizeof(*list));
  int i, k;
  double d, *p, *q, *r, *v0;

  assert(x->nact == y->nact && x->nact <= n0);

  for (i = 0, k = 0; i < x->nvec; i++)
    if (mask == NULL || mask[i] != 0)
      list[k++] = i;
  assert (k == x->nact);

  r = v0 = malloc(x->nact * sizeof(*v0));
  for (i = 0; i < x->nact; i++) {
    p = y->data + y->list[i] * nlocal;
    q = x->data + x->list[i] * nlocal;
    d = 0.0;
    for (k = 0; k < nlocal; k++)
      d += *(p++) * *(q++);
    *(r++) = d;
  }

  if (nprocs > 1) {
    r = malloc(x->nact * sizeof(*r));
    MPI_Allreduce(v0, r, x->nact, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    free(v0);
    v0 = r;
  }

  for (i = 0, r = v0; i < x->nact; i++)
    diag[list[i]] = *(r++);

  free(v0);
  free(list);
  CHECK(x, y);
}

static void
MultiVectorByDiagonal(void *x_, int *mask, int n0, double *alpha, void *y_)
/* y(<y_mask>) = alpha(<mask>) .* x(<x_mask>) */
{
  VEC *x = x_, *y = y_;
  int *list = malloc(n0 * sizeof(*list));
  int i, k;
  double d, *p, *q;

  assert(x->nact == y->nact && x->nact <= n0);

  for (i = 0, k = 0; i < x->nvec; i++)
    if (mask == NULL || mask[i] != 0)
      list[k++] = i;
  assert (k == x->nact);

  for (i = 0; i < x->nact; i++) {
    p = y->data + y->list[i] * nlocal;
    q = x->data + x->list[i] * nlocal;
    d = alpha[list[i]];
    for (k = 0; k < nlocal; k++)
      *(p++) = *(q++) * d;
  }

  free(list);
  CHECK(x, y);
}

static void 
MultiVectorByMatrix(void *x_, int gh, int h, int w, double *v, void *y_)
{
  VEC *x = x_, *y = y_;
  int i, j, k;
  double d, *p, *q;

  assert (h == x->nact && w == y->nact);
 
  for (j = 0; j < w; j++) {
    p = y->data + y->list[j] * nlocal;
    q = x->data + x->list[0] * nlocal;
    d = *(v++);
    for (k = 0; k < nlocal; k++)
      p[k] = d * *(q++);
    for (i = 1; i < h; i++) {
      q = x->data + x->list[i] * nlocal;
      d = *(v++);
      for (k = 0; k < nlocal; k++)
        p[k] += d * *(q++);
    }
    v += gh - h;
  }

  CHECK(x, y);
}

static void
MultiVectorAxpy(double alpha, void *x_, void *y_)
{
  VEC *x = x_, *y = y_;
  int i, j;
  double *p, *q;

  assert (x->nact == y->nact);
  for (i = 0; i < x->nact; i++) {
    p = y->data + y->list[i] * nlocal;
    q = x->data + x->list[i] * nlocal;
    for (j = 0; j < nlocal; j++)
      *(p++) += alpha * *(q++);
  }
  CHECK(x, y);
}

static void
MatMultiVec(void *data_, void *x_, void *y_)
{
  VEC *x = x_, *y = y_;
  void (*func)(double *x, double *y) = data_;
  int i;

  assert (x->nact == y->nact);
  for (i = 0; i < x->nact; i++) {
    func(x->data + x->list[i] * nlocal, y->data + y->list[i] * nlocal);
  }
  CHECK(x, y);
}

int
main(int argc, char **argv)
{
  VEC *x0;
  mv_MultiVectorPtr x;
  double *eigs;
  double *resid;
  int i, it;
  lobpcg_Tolerance lobpcg_tol;
  mv_InterfaceInterpreter ii;
  lobpcg_BLASLAPACKFunctions blap_fn;

  /* init interface interpreter */
  /* Vector part */
  ii.CreateVector = NULL;
  ii.DestroyVector = NULL;
  ii.InnerProd = NULL; 
  ii.CopyVector = NULL;
  ii.ClearVector = NULL;
  ii.SetRandomValues = NULL;
  ii.ScaleVector = NULL;
  ii.Axpy = NULL;

  /* Multivector part */
  ii.CopyCreateMultiVector = CreateCopyMultiVector;
  ii.DestroyMultiVector = DestroyMultiVector;

  ii.Width = MultiVectorWidth;
  ii.Height = NULL;
  ii.SetMask = MultiSetMask;
  ii.CopyMultiVector = CopyMultiVector;
  ii.ClearMultiVector = ClearMultiVector;
  ii.SetRandomVectors = SetMultiVectorRandomValues;
  ii.MultiInnerProd = MultiInnerProd;
  ii.MultiInnerProdDiag = MultiInnerProdDiag;
  ii.MultiVecMat = MultiVectorByMatrix;
  ii.MultiVecMatDiag = MultiVectorByDiagonal;
  ii.MultiAxpy = MultiVectorAxpy;
  ii.MultiXapy = NULL;
  ii.Eval = NULL;
  
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);

  /* compute local vector size */
  nlocal = nglobal / nprocs + (myid < nglobal % nprocs ? 1 : 0);
  prev = (myid == 0 ? MPI_PROC_NULL : myid - 1);
  next = (myid == nprocs - 1 ? MPI_PROC_NULL : myid + 1);

  eigs = (double *)malloc(neig * sizeof(double));
  resid = (double *)malloc(neig * sizeof(double));
    
  x0 = CreateMultiVector(neig);
  SetMultiVectorRandomValues(x0, 1);
  x = mv_MultiVectorWrap(&ii, x0, 1/*ownsData, FIXME: what's its meaning?*/);

  lobpcg_tol.absolute = 1e-10;
  lobpcg_tol.relative = 1e-50;

  blap_fn.dpotrf = dpotrf_;
  blap_fn.dsygv = dsygv_;
  lobpcg_solve( x,		/* eigenvectors */
		matveca,	/* A data (function pointer here) */
		MatMultiVec,	/* A*x */
		matvecb,	/* B data (function pointer here) */
		MatMultiVec,	/* B*x */
		NULL,		/* T data */
		NULL,		/* T*x */
		NULL,		/* blockVectorY (constraints) */
		blap_fn,	/* BLAS/LAPACK functions */
		lobpcg_tol,	/* tolerances */
		400,		/* MAXIT */
		0,		/* verbose level */
		&it,		/* iteration nunmber */
		eigs,		/* lambda's */
		NULL,		/* lambda history: neig * (it + 1) */
		0,		/* lamda history LDA */
		resid,		/* residual norms */
		NULL,		/* residual norms history: neig * (it + 1) */
		0		/* residual norms history LDA */
  );

  if (myid == 0) {
    printf("%d iterations.\n", it);
    it = nglobal + 1;
    for (i = 0; i < neig; i++) {
      double lambda = 2.*(1.-cos((i + 1) * M_PI / it)) * it * it;
      printf("eigenvalue %d = %lg, resid = %lg, error = %lg\n",
		i + 1, eigs[i], resid[i], eigs[i] - lambda);
    }
  }

  free(eigs);
  free(resid);
  DestroyMultiVector(x0);
  MPI_Finalize();

  return 0;
}
